using SphericalHarmonicArrays,OffsetArrays
import SphericalHarmonicArrays: SizeMismatchArrayModeError, 
SizeMismatchError, MismatchedDimsError, UnexpectedAxisTypeError,
NotAnSHAxisError

using Test

@testset "Constructors" begin
	mode_lm = LM(0:1,0:0)
	mode_ml = ML(0:1,0:0)
	mode_l₂l₁ = L₂L₁Δ(0:2,2,0:2)
	@testset "0dim" begin
	    @test SHArray(zeros()) == zeros()
	end
	@testset "normal" begin
	    arr = zeros(ComplexF64,length(mode_lm))
	    @test SHArray(arr) == arr
	end
	@testset "one SH axis" begin
		arr = zeros(ComplexF64,length(mode_lm))
		@testset "SHArray" begin
		    @test SHArray(mode_lm) == arr
		    @test SHArray((mode_lm,)) == arr
		    @test SHArray(mode_lm,(1,)) == arr
		    @test SHArray((mode_lm,),(1,)) == arr
		    @test SHArray(arr,mode_lm) == arr
		    @test SHArray(arr,(mode_lm,)) == arr
		    @test SHArray(arr,mode_lm,(1,)) == arr
		    @test SHArray(arr,(mode_lm,),(1,)) == arr
		end
		@testset "SHVector" begin
		    @test SHVector(mode_lm) == arr
		    @test SHVector{ComplexF64}(mode_lm) == arr
		    @test SHVector{Float64}(mode_lm) == real(arr)
		    @test SHVector(arr,mode_lm) == arr
		end

	    @testset "Errors" begin
	    	@test_throws UnexpectedAxisTypeError SHArray(mode_lm,(2,))
	    end
	end
    @testset "2D mixed axes" begin
    	@testset "first SH" begin
		    arr = zeros(ComplexF64,length(mode_lm),2)
		    ax = (mode_lm,1:2)
		    shdims = (1,)
		    @test SHArray(ax) == arr
		    @test SHArray{ComplexF64}(ax) == arr
		    @test SHArray{Float64}(ax) == real(arr)
		    @test SHArray(ax,shdims) == arr
		    @test SHArray{ComplexF64}(ax,shdims) == arr
		    @test SHArray{Float64}(ax,shdims) == real(arr)
		    @test SHArray(arr,ax) == arr
		    @test SHArray(arr,ax,shdims) == arr

		    @testset "Errors" begin
		    	@test_throws MismatchedDimsError SHArray(ax,(1,2,3))
		    	@test_throws UnexpectedAxisTypeError SHArray(ax,(2,))
		    end
		end
		@testset "second SH" begin
			arr = zeros(ComplexF64,2,length(mode_lm),)
			ax = (1:2,mode_lm)
			shdims = (2,)
		    @test SHArray(ax) == arr
		    @test SHArray{ComplexF64}(ax) == arr
		    @test SHArray{Float64}(ax) == real(arr)
		    @test SHArray(ax,shdims) == arr
		    @test SHArray{ComplexF64}(ax,shdims) == arr
		    @test SHArray{Float64}(ax,shdims) == real(arr)
		    @test SHArray(arr,ax) == arr
		    @test SHArray(arr,ax,shdims) == arr
		    
		    @testset "Errors" begin
		    	@test_throws MismatchedDimsError SHArray(ax,(1,2,3))
		    	@test_throws UnexpectedAxisTypeError SHArray(ax,(1,))
		    	@test_throws SizeMismatchArrayModeError SHArray(arr,(1:2,LM(l_range(mode_lm))),(2,))
		    	@test_throws SizeMismatchArrayModeError SHArray(arr,(1:2,LM(l_range(mode_lm))))
		    	@test_throws SizeMismatchError SHArray(arr,(1:3,LM(l_range(mode_lm))),(2,))
		    	@test_throws SizeMismatchError SHArray(arr,(1:3,LM(l_range(mode_lm))))
		    end
		end
		@testset "both SH" begin
			arr = zeros(ComplexF64,length(mode_lm),length(mode_lm))
			ax = (mode_lm,mode_lm)
			shdims = (1,2)
		    @test SHArray(ax) == arr
		    @test SHArray{ComplexF64}(ax) == arr
		    @test SHArray{Float64}(ax) == real(arr)
		    @test SHArray(ax,shdims) == arr
		    @test SHArray{ComplexF64}(ax,shdims) == arr
		    @test SHArray{Float64}(ax,shdims) == real(arr)
		    @test SHArray(arr,ax) == arr
		    @test SHArray(arr,ax,shdims) == arr

		    @testset "Errors" begin
		    	@test_throws MismatchedDimsError SHArray(ax,(1,2,3))
		    	@test_throws UnexpectedAxisTypeError SHArray{Float64}(ax,(1,))
		    	@test_throws UnexpectedAxisTypeError SHArray{Float64}(ax,(2,))
		    	ax = (LM(l_range(mode_lm)),LM(l_range(mode_lm)))
		    	@test_throws SizeMismatchArrayModeError SHArray(arr,ax,(1,2))
		    	@test_throws SizeMismatchArrayModeError SHArray(arr,ax)
		    end
		end
		@testset "SHMatrix" begin
			function testSHMatrix(mode1,mode2)
				arr = zeros(ComplexF64,length(mode1),length(mode2))
				@test SHMatrix(arr,mode1,mode2) isa SHMatrix
			    @test SHMatrix(mode1,mode2) isa SHMatrix
			    @test SHMatrix((mode1,mode2)) isa SHMatrix
			    @test SHMatrix(arr,mode1,mode2) == arr
			    @test SHMatrix(arr,(mode1,mode2)) == arr
			    @test SHMatrix(mode1,mode2) == arr
			    @test SHMatrix((mode1,mode2)) == arr
			    @test SHMatrix{Float64}(mode1,mode2) == real(arr)
			    @test SHMatrix{Float64}((mode1,mode2)) == real(arr)
			    @test shdims(SHMatrix(arr,mode1,mode2)) == (1,2)
			    @test shdims(SHMatrix(arr,(mode1,mode2))) == (1,2)
			    @test shdims(SHMatrix(mode1,mode2)) == (1,2)
			    @test shdims(SHMatrix((mode1,mode2))) == (1,2)
			    @test modes(SHMatrix(arr,mode1,mode2)) == (mode1,mode2)
			    @test modes(SHMatrix(arr,(mode1,mode2))) == (mode1,mode2)
			    @test modes(SHMatrix(mode1,mode2)) == (mode1,mode2)
			    @test modes(SHMatrix((mode1,mode2))) == (mode1,mode2)	
			end
			@testset "LM LM" begin
			    testSHMatrix(mode_lm,mode_lm)
			end
			@testset "LM ML" begin
			    testSHMatrix(mode_lm,mode_ml)
			end
			@testset "ML LM" begin
			    testSHMatrix(mode_ml,mode_lm)
			end
			@testset "ML ML" begin
			    testSHMatrix(mode_ml,mode_ml)
			end
		    @testset "LM L₂L₁Δ" begin
			    testSHMatrix(mode_lm,mode_l₂l₁)
			end
			@testset "ML L₂L₁Δ" begin
			    testSHMatrix(mode_ml,mode_l₂l₁)
			end
		    @testset "L₂L₁Δ LM" begin
			    testSHMatrix(mode_l₂l₁,mode_lm)
			end
			@testset "L₂L₁Δ ML" begin
			    testSHMatrix(mode_l₂l₁,mode_ml)
			end
			@testset "L₂L₁Δ L₂L₁Δ" begin
			    testSHMatrix(mode_l₂l₁,mode_l₂l₁)
			end
		end
	end
end

@testset "Indexing" begin
	mode_lm = LM(0:1,0:0)
	@testset "1D SH" begin
	    v = SHVector(mode_lm)
	    @testset "getindex" begin
	    	@testset "linear" begin
		    	for i in eachindex(v)
			    	@test v[i] == 0
			    end
	    	end
	    	@testset "mode_lm" begin
			    for m in mode_lm
			    	@test v[m] == 0
			    end
	    	end
	    end
	    @testset "setindex" begin
	    	@testset "linear" begin
			    for i in eachindex(v)
			    	v[i] = i
			    	@test v[i] == i
			    end
	    	end
	    	@testset "mode_lm" begin
			    for (i,m) in enumerate(mode_lm)
			    	v[m] = i
			    	@test v[m] == i
			    end
	    	end
	    end
	end
	@testset "1D normal" begin
	    v = SHArray(zeros(length(mode_lm)))
	    @testset "getindex" begin
	    	@testset "linear" begin
		    	for i in eachindex(v)
			    	@test v[i] == 0
			    end
			end
			@testset "mode_lm" begin
			    for m in mode_lm
			    	@test_throws NotAnSHAxisError v[m]
			    end
	    	end
	    end
	    @testset "setindex" begin
	    	@testset "linear" begin
			    for i in eachindex(v)
			    	v[i] = i
			    	@test v[i] == i
			    end
	    	end
	    	@testset "mode_lm" begin
			    for (i,m) in enumerate(mode_lm)
			    	@test_throws NotAnSHAxisError v[m] = i
			    end
	    	end
	    end
	end
	@testset "2D mixed axes" begin
		@testset "first SH" begin
		    v = SHArray((mode_lm,1:2))
		    @testset "getindex" begin
		    	@testset "linear" begin
			    	for i in eachindex(v)
				    	@test v[i] == 0
				    end
		    	end
		    	@testset "mode_lm" begin
				    for ind2 in axes(v,2), m in mode_lm
				    	@test v[m,ind2] == 0
				    end
		    	end
		    end
		    @testset "setindex" begin
		    	@testset "linear" begin
				    for i in eachindex(v)
				    	v[i] = i
				    	@test v[i] == i
				    end
		    	end
		    	@testset "mode_lm" begin
				    for ind2 in axes(v,2),(i,m) in enumerate(mode_lm)
				    	v[m,ind2] = i
				    	@test v[m,ind2] == i
				    end
		    	end
		    end
		end
		@testset "second SH" begin
		    v = SHArray((1:2,mode_lm))
		    @testset "getindex" begin
		    	@testset "linear" begin
			    	for i in eachindex(v)
				    	@test v[i] == 0
				    end
		    	end
		    	@testset "mode_lm" begin
				    for (i,m) in enumerate(mode_lm),ind1 in axes(v,1)
				    	@test v[ind1,m] == 0
				    end
		    	end
		    end
		    @testset "setindex" begin
		    	@testset "linear" begin
				    for i in eachindex(v)
				    	v[i] = i
				    	@test v[i] == i
				    end
		    	end
		    	@testset "mode_lm" begin
				    for (i,m) in enumerate(mode_lm),ind1 in axes(v,1)
				    	v[ind1,m] = i
				    	@test v[ind1,m] == i
				    end
		    	end
		    end
		end
		@testset "both SH" begin
		    v = SHArray((mode_lm,mode_lm))
		    @testset "getindex" begin
		    	@testset "linear" begin
			    	for i in eachindex(v)
				    	@test v[i] == 0
				    end
		    	end
		    	@testset "mode_lm" begin
				    for m2 in mode_lm,m1 in mode_lm
				    	@test v[m1,m2] == 0
				    end
		    	end
		    end
		    @testset "setindex" begin
		    	@testset "linear" begin
				    for i in eachindex(v)
				    	v[i] = i
				    	@test v[i] == i
				    end
		    	end
		    	@testset "mode_lm" begin
				    for (i,m2) in enumerate(mode_lm),m1 in mode_lm
				    	v[m1,m2] = i
				    	@test v[m1,m2] == i
				    end
		    	end
		    end
		end
	end
end

@testset "similar" begin
    mode_lm = LM(0:1,0:0)
	mode_ml = ML(0:1,0:0)
	mode_l₂l₁ = L₂L₁Δ(0:2,2,0:2)

	function testsimilar(arr)
		@test similar(arr) isa typeof(arr)
		@test size(parent(similar(arr))) == size(parent(arr))
		@test modes(similar(arr)) == modes(arr)
		@test shdims(similar(arr)) == shdims(arr)
	end
	
	SA = SHArray((mode_lm,1:2,mode_ml))
	testsimilar(SA)
	v = SHVector(mode_lm)
	testsimilar(v)
	M = SHMatrix(mode_lm,mode_lm)
	testsimilar(M)
end

@testset "nested" begin
	sharr = SHVector(LM(0:2))
	v = [sharr,sharr]
	n = length(v)
    @test typeof(v) == Array{typeof(sharr),1}
    shv = SHVector(v,LM(n:n,1:n)) # get the number of modes to match
    @test typeof(shv) == SHArray{typeof(sharr),1,typeof(v),Tuple{LM},1}
    @test_throws SizeMismatchArrayModeError SHVector(v,LM(0:0,0:0))
end